import base64
import logging
import urllib
import urllib.parse
from dataclasses import dataclass
from typing import Dict, List, Optional
from uuid import uuid4

import requests
from label_studio_ml.model import LabelStudioMLBase
from label_studio_ml.response import ModelResponse
from label_studio_sdk.label_interface.objects import PredictionValue

from label_studio_paddleocr.config import settings

logger = logging.getLogger(__name__)


@dataclass
class Block:
    """A text block detected by OCR."""

    label: str
    content: str
    x1: int
    y1: int
    x2: int
    y2: int
    width: int
    height: int


class PaddleOCR(LabelStudioMLBase):
    """Use PaddleOCR to extract text from images."""

    def setup(self):
        # Get configuration for RectangleLabels and TextArea
        self.rl_from_name, self.rl_to_name, _ = self.get_first_tag_occurence(
            "RectangleLabels", "Image"
        )
        self.ta_from_name, self.ta_to_name, _ = self.get_first_tag_occurence(
            "TextArea", "Image"
        )

    def predict(
        self, tasks: List[Dict], context: Optional[Dict] = None, **kwargs
    ) -> ModelResponse:
        """Write the inference logic here
        :param tasks: [Label Studio tasks in JSON format](https://labelstud.io/guide/task_format.html)
        :param context: [Label Studio context in JSON format](https://labelstud.io/guide/ml_create#Implement-prediction-logic)
        :return model_response
            ModelResponse(predictions=predictions) with
            predictions: [Predictions array in JSON format](https://labelstud.io/guide/export.html#Label-Studio-JSON-format-of-annotated-tasks)
        """
        logger.debug(f"Run prediction on {tasks}")

        predictions = []
        for task in tasks:
            task_id = task["id"]
            task_image = task["data"]["image"]
            logger.info(f"Processing task {task_id} with image {task_image}")

            image_local_path = self.get_local_path(
                task_image,
                task_id=task_id,
                ls_host=settings.label_studio_url,
                ls_access_token=settings.label_studio_api_key,
            )

            blocks = self._ocr(image_local_path)
            prediction = self._generate_predications(blocks)
            predictions.append(prediction)

        return ModelResponse(
            model_version=settings.model_version, predictions=predictions
        )

    def _ocr(self, image_path: str) -> list[Block]:
        """Send image to PaddleOCR and handle OCR results
        :param image_path: Local path to the image file
        :return: OCR results as a dictionary
        """

        # https://www.paddleocr.ai/main/version3.x/pipeline_usage/PaddleOCR-VL.html#43
        with open(image_path, "rb") as file:
            image_bytes = file.read()
            image_data = base64.b64encode(image_bytes).decode("ascii")

        payload = {
            "file": image_data,
            "fileType": 1,  # 0 for pdf, 1 for image
        }
        url = urllib.parse.urljoin(settings.paddleocr_url, "/layout-parsing")
        logger.info(f"Sending request to {url}")
        response = requests.post(url, json=payload)

        # handle response
        blocks = []
        if response.status_code != 200:
            logger.warning(
                f"Request failed with status {response.status_code}: {response.text}"
            )
            return blocks

        for layout_parsing_result in response.json()["result"]["layoutParsingResults"]:
            pruned_result = layout_parsing_result["prunedResult"]
            width = pruned_result["width"]
            height = pruned_result["height"]
            for res in pruned_result["parsing_res_list"]:
                bbox = res["block_bbox"]
                block = Block(
                    res["block_label"],
                    res["block_content"],
                    bbox[0],
                    bbox[1],
                    bbox[2],
                    bbox[3],
                    width=width,
                    height=height,
                )
                blocks.append(block)
        return blocks

    def _generate_predications(self, blocks: list[Block]) -> PredictionValue:
        """Generate predictions from OCR blocks
        :param blocks: List of OCR blocks
        :return: SingleTaskPredictions object
        """
        result = []
        for block in blocks:
            # https://github.com/PaddlePaddle/PaddleX/blob/37f1ffdc1daae40bbc5e17141d064951be9796ed/paddlex/inference/pipelines/paddleocr_vl/pipeline.py#L259
            category = "OCR:"
            if block.label == "table":
                category = "Table Recognition:"
            elif block.label == "chart":
                category = "Chart Recognition:"
            elif "formula" in block.label and block.label != "formula_number":
                category = "Formula Recognition:"

            rectangle_sizes = {
                "x": block.x1 / block.width * 100,
                "y": block.y1 / block.height * 100,
                "width": (block.x2 - block.x1) / block.width * 100,
                "height": (block.y2 - block.y1) / block.height * 100,
            }
            id = str(uuid4())

            result.append(
                {
                    "id": id,
                    "from_name": self.rl_from_name,
                    "to_name": self.rl_to_name,
                    "type": "rectanglelabels",
                    "value": {
                        **rectangle_sizes,
                        "rectanglelabels": [category],
                    },
                }
            )
            result.append(
                {
                    "id": id,
                    "from_name": self.ta_from_name,
                    "to_name": self.ta_to_name,
                    "type": "textarea",
                    "value": {
                        **rectangle_sizes,
                        "text": [block.content],
                    },
                }
            )
        return PredictionValue(result=result)



from typing import Optional
from pydantic_settings import (
    BaseSettings,
    PydanticBaseSettingsSource,
    SettingsConfigDict,
    YamlConfigSettingsSource,
)


class AppSettings(BaseSettings):
    """Configuration settings for the application."""

    model_version: str = "PaddleOCR-VL"
    
    paddleocr_url: str

    label_studio_url: Optional[str] = None
    label_studio_api_key: Optional[str] = None

    model_config = SettingsConfigDict(yaml_file="app.yaml")

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: type[BaseSettings],
        init_settings: PydanticBaseSettingsSource,
        env_settings: PydanticBaseSettingsSource,
        dotenv_settings: PydanticBaseSettingsSource,
        file_secret_settings: PydanticBaseSettingsSource,
    ) -> tuple[PydanticBaseSettingsSource, ...]:
        return (
            init_settings,
            YamlConfigSettingsSource(settings_cls),
            env_settings,
            dotenv_settings,
            file_secret_settings,
        )


settings = AppSettings()



import argparse
import logging
import os
import sys

from label_studio_ml.api import init_app

from label_studio_paddleocr.model import PaddleOCR


def init_logging():
    logging.basicConfig(
        level=logging.INFO,
        stream=sys.stdout,
        format="%(asctime)s %(levelname)-5s [%(name)s::%(funcName)s::%(lineno)d] : %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )


def parse_args():
    parser = argparse.ArgumentParser(description="Label studio PaddleOCR")
    parser.add_argument(
        "-p", "--port", dest="port", type=int, default=9090, help="Server port"
    )
    parser.add_argument(
        "--host", dest="host", type=str, default="0.0.0.0", help="Server host"
    )
    parser.add_argument(
        "--kwargs",
        "--with",
        dest="kwargs",
        metavar="KEY=VAL",
        nargs="+",
        type=lambda kv: kv.split("="),
        help="Additional LabelStudioMLBase model initialization kwargs",
    )
    parser.add_argument(
        "-d", "--debug", dest="debug", action="store_true", help="Switch debug mode"
    )
    parser.add_argument(
        "--log-level",
        dest="log_level",
        choices=["DEBUG", "INFO", "WARNING", "ERROR"],
        default="INFO",
        help="Logging level",
    )
    parser.add_argument(
        "--model-dir",
        dest="model_dir",
        default=os.path.dirname(__file__),
        help="Directory where models are stored (relative to the project directory)",
    )
    parser.add_argument(
        "--check",
        dest="check",
        action="store_true",
        help="Validate model instance before launching server",
    )
    parser.add_argument(
        "--basic-auth-user",
        default=os.environ.get("ML_SERVER_BASIC_AUTH_USER", None),
        help="Basic auth user",
    )
    parser.add_argument(
        "--basic-auth-pass",
        default=os.environ.get("ML_SERVER_BASIC_AUTH_PASS", None),
        help="Basic auth pass",
    )
    return parser.parse_args()


def main():
    init_logging()

    args = parse_args()

    # setup logging level
    if args.log_level:
        logging.root.setLevel(args.log_level)

    def isfloat(value):
        try:
            float(value)
            return True
        except ValueError:
            return False

    def parse_kwargs():
        param = dict()
        for k, v in args.kwargs:
            if v.isdigit():
                param[k] = int(v)
            elif v == "True" or v == "true":
                param[k] = True
            elif v == "False" or v == "false":
                param[k] = False
            elif isfloat(v):
                param[k] = float(v)
            else:
                param[k] = v
        return param

    kwargs = dict()

    if args.kwargs:
        kwargs.update(parse_kwargs())

    if args.check:
        print('Check "' + PaddleOCR.__name__ + '" instance creation..')
        PaddleOCR(**kwargs)

    app = init_app(
        model_class=PaddleOCR,
        basic_auth_user=args.basic_auth_user,
        basic_auth_pass=args.basic_auth_pass,
    )

    app.run(host=args.host, port=args.port, debug=args.debug)


if __name__ == "__main__":
    main()
else:
    # for uWSGI use
    app = init_app(model_class=PaddleOCR)



import sys

from gunicorn.app.wsgiapp import run


def main():
    # uv run label-studio-paddleocr -- --workers 4 --log-level debug
    sys.argv = [
        "gunicorn",
        *sys.argv[1:],
        "label_studio_paddleocr._wsgi:app",
    ]

    run()


if __name__ == "__main__":
    main()
